@@ -21,13 +21,24 @@ do
 	*)      usage; exit 1 ;;
   esac
 done
+fixout() {
+  sed -e '/^220 local.host mailfront ESMTP/d; s/\[[0-9]*\]:/[#]:/;'
+}
+fixtls() {
+  sed -e '
+    s/\[[0-9]*\]:/[#]:/;
+    /INTERNAL ERROR: Uninitialized client session cache/d;
+    s/^.* \(LOG[1-9]\)/\1/;
+    /^LOG[1-9]\[ui\]: /d;
+    s/handshake: TLS[1-9]\.[0-9]/handshake: TLS#.#/;
+  '
+}
+
 smtpfront() {
-  $src/mailfront smtp "$@" 2>&1 \
-  | sed -e '/^220 local.host mailfront ESMTP/d; s/\[[0123456789]*\]:/[#]:/'
+  $src/mailfront smtp "$@" 2>&1 | fixout
 }
 sfecho() {
-  $src/mailfront smtp echo "$@" 2>/dev/null \
-  | grep -v '^220 local.host mailfront ESMTP'
+  $src/mailfront smtp echo "$@" 2>/dev/null | fixout
 }
 pfauth() {
   $src/pop3front-auth "$@" echo Yes. 2>/dev/null \
@@ -78,6 +89,41 @@ cleanup_queuedir() {
     rm -rf $QUEUEDIR
 }
 
+setup_cert() {
+    certfile=$tmp/tlscrt.pem
+    keyfile=$tmp/tlskey.pem
+    dhparams=$tmp/dhparams.pem
+
+    if ! [ -s $certfile ]
+    then
+        certtool --generate-privkey --sec-param low --outfile $keyfile 2> /dev/null
+        template=$tmp/cert.cfg
+        echo 'cn = example.com' > $template
+        certtool --generate-self-signed --load-privkey $keyfile --template $template --outfile $certfile 2> /dev/null
+        rm -f $template
+        certtool --generate-dh-params --sec-param low --outfile $dhparams 2> /dev/null
+    fi
+}
+
+make_stunnel_conf() {
+    modules="$1"
+    shift
+    {
+        cat <<EOF
+client = yes
+debug = notice
+exec = $src/mailfront
+execArgs = mailfront $modules
+foreground = yes
+syslog = no
+EOF
+        for line in "$@"
+        do
+            echo "$line"
+        done
+    } > $tmp/stunnel.conf
+}
+
 PROTO=TEST
 TESTLOCALIP=1.2.3.4
 TESTREMOTEIP=5.6.7.8
@@ -187,53 +233,6 @@ run_compare_test tests/patterns-after  <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
-##### Test tests/smtpfront-bad-bounce #####
-
-runtest() {
-# Note: this test no longer tests anything significant.
-
-sfecho <<EOF
-MAIL FROM:<notbounce@example.com>
-RCPT TO:<addr1@example.com>
-RCPT TO:<addr2@example.com>
-DATA
-.
-EOF
-
-sfecho <<EOF
-MAIL FROM:<>
-RCPT TO:<addr1@example.com>
-DATA
-.
-EOF
-
-sfecho <<EOF
-MAIL FROM:<>
-RCPT TO:<addr1@example.com>
-RCPT TO:<addr2@example.com>
-DATA
-.
-EOF
-}
-vecho "Running test tests/smtpfront-bad-bounce "
-run_compare_test tests/smtpfront-bad-bounce  <<END_OF_TEST_RESULTS
-250 Sender='notbounce@example.com'.^M
-250 Recipient='addr1@example.com'.^M
-250 Recipient='addr2@example.com'.^M
-354 End your message with a period on a line by itself.^M
-250 Received 0 bytes.^M
-250 Sender=''.^M
-250 Recipient='addr1@example.com'.^M
-354 End your message with a period on a line by itself.^M
-250 Received 0 bytes.^M
-250 Sender=''.^M
-250 Recipient='addr1@example.com'.^M
-250 Recipient='addr2@example.com'.^M
-354 End your message with a period on a line by itself.^M
-250 Received 0 bytes.^M
-END_OF_TEST_RESULTS
-
-
 ##### Test tests/plugin-cvm-auth-caps #####
 
 runtest() {
@@ -406,6 +405,19 @@ run_compare_test tests/plugin-reject '-rej' <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
+##### Test tests/qmtp-echo #####
+
+runtest() {
+printf '6:\ntest\n,13:a@example.com,17:13:b@example.com,,' \
+| $src/mailfront qmtp echo 2>/dev/null
+echo
+}
+vecho "Running test tests/qmtp-echo "
+run_compare_test tests/qmtp-echo  <<END_OF_TEST_RESULTS
+18:KReceived 5 bytes.,
+END_OF_TEST_RESULTS
+
+
 ##### Test tests/plugin-counters-looping-received #####
 
 runtest() {
@@ -503,22 +515,73 @@ run_compare_test tests/pop3front-maildir-size  <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
-##### Test tests/smtpfront-quotes #####
+##### Test tests/starttls-works #####
 
 runtest() {
-sfecho <<EOF
-MAIL FROM:<"me, myself, and I"@example.net>
-RCPT TO:<"you, yourself, and you"@example.com>
-RCPT TO:<him\,himself@example.com>
-RCPT TO:<@somewhere,@elsewhere:two@example.com>
-EOF
+setup_cert
+make_stunnel_conf 'smtp echo accept' 'protocol = smtp'
+
+{
+    # Add sleeps between echos to avoid races
+    echo 'MAIL FROM:<nobody>'
+    sleep 1
+    echo 'RCPT TO:<somebody>'
+    sleep 1
+} | env TLS_CERTFILE=$certfile TLS_KEYFILE=$keyfile \
+stunnel $tmp/stunnel.conf 2>&1 | fixtls
 }
-vecho "Running test tests/smtpfront-quotes "
-run_compare_test tests/smtpfront-quotes  <<END_OF_TEST_RESULTS
-250 Sender='me, myself, and I@example.net'.^M
-250 Recipient='you, yourself, and you@example.com'.^M
-250 Recipient='him,himself@example.com'.^M
-250 Recipient='two@example.com'.^M
+vecho "Running test tests/starttls-works "
+run_compare_test tests/starttls-works  <<END_OF_TEST_RESULTS
+LOG5[#]: Service [stunnel] accepted connection
+220 local.host mailfront ESMTP^M
+mailfront[#]: Starting TLS handshake
+mailfront[#]: TLS handshake: TLS#.# ECDHE-RSA AES-256-GCM AEAD
+mailfront[#]: MAIL FROM:<nobody>
+250 Sender='nobody'.^M
+mailfront[#]: RCPT TO:<somebody>
+250 Recipient='somebody'.^M
+LOG5[#]: Connection closed: 38 byte(s) sent to TLS, 49 byte(s) sent to socket
+mailfront[#]: bytes in: 38 bytes out: 49
+END_OF_TEST_RESULTS
+
+
+##### Test tests/plugin-rbl-whitelist #####
+
+runtest() {
+local enabled="$1"
+
+export RBL_BLACKLISTS=zen.spamhaus.org
+if $enabled
+then
+    export RBL_WHITELISTS=zen.spamhaus.org
+fi
+export RBL_DEBUG=1
+export TCPREMOTEIP=127.0.0.2
+
+smtpfront echo rbl accept <<EOF
+HELO me
+MAIL FROM:<somewhere>
+EOF
+
+unset RBL_BLACKLISTS RBL_DEBUG TCPLOCALIP
+}
+vecho "Running test tests/plugin-rbl-whitelist 'false'"
+run_compare_test tests/plugin-rbl-whitelist 'false' <<END_OF_TEST_RESULTS
+mailfront[#]: rbl: blacklisted by zen.spamhaus.org: https://www.spamhaus.org/sbl/query/SBL2 // https://www.spamhaus.org/query/ip/127.0.0.2
+250 local.host^M
+mailfront[#]: MAIL FROM:<somewhere>
+mailfront[#]: 451 Blocked: https://www.spamhaus.org/sbl/query/SBL2
+451 Blocked: https://www.spamhaus.org/sbl/query/SBL2^M
+mailfront[#]: bytes in: 30 bytes out: 102
+END_OF_TEST_RESULTS
+
+vecho "Running test tests/plugin-rbl-whitelist 'true'"
+run_compare_test tests/plugin-rbl-whitelist 'true' <<END_OF_TEST_RESULTS
+mailfront[#]: rbl: whitelisted by zen.spamhaus.org: https://www.spamhaus.org/sbl/query/SBL2 // https://www.spamhaus.org/query/ip/127.0.0.2
+250 local.host^M
+mailfront[#]: MAIL FROM:<somewhere>
+250 Sender='somewhere'.^M
+mailfront[#]: bytes in: 30 bytes out: 73
 END_OF_TEST_RESULTS
 
 
@@ -626,109 +689,105 @@ END_OF_TEST_RESULTS
 
 ##### Test tests/plugin-rbl #####
 
-#runtest() {
-#local ip="$1"
-#local queuedir="$2"
-#
-#export RBL_BLACKLISTS=sbl-xbl.spamhaus.org
-#export RBL_DEBUG=1
-#export TCPREMOTEIP=$ip
-#
-#setup_queuedir
-#if $queuedir
-#then
-#    export RBL_QUEUEDIR=$QUEUEDIR
-#fi
-#
-#smtpfront echo rbl accept <<EOF
-#HELO me
-#MAIL FROM:<somewhere>
-#RCPT TO:<elsewhere>
-#DATA
-#Subject: test
-#.
-#EOF
-#
-#dump_queuedir
-#
-#cleanup_queuedir
-#
-#unset RBL_BLACKLISTS RBL_DEBUG RBL_QUEUEDIR TCPLOCALIP QUEUEDIR
-#rm -rf $tmp/queuedir $tmp/queuedir
-#}
-#vecho "Running test tests/plugin-rbl '127.0.0.2' 'false'"
-#run_compare_test tests/plugin-rbl '127.0.0.2' 'false' <<END_OF_TEST_RESULTS
-#mailfront[#]: rbl: sbl-xbl.spamhaus.org:
-#mailfront[#]: http://www.spamhaus.org/sbl/query/SBL233
-#mailfront[#]: http://www.spamhaus.org/query/bl?ip=127.0.0.2
-#250 local.host^M
-#mailfront[#]: MAIL FROM:<somewhere>
-#mailfront[#]: 451 Blocked: http://www.spamhaus.org/sbl/query/SBL233
-#451 Blocked: http://www.spamhaus.org/sbl/query/SBL233^M
-#mailfront[#]: RCPT TO:<elsewhere>
-#mailfront[#]: 503 5.5.1 You must send MAIL FROM: first
-#503 5.5.1 You must send MAIL FROM: first^M
-#mailfront[#]: 503 5.5.1 You must send MAIL FROM: first
-#503 5.5.1 You must send MAIL FROM: first^M
-#mailfront[#]: Subject: test
-#mailfront[#]: 500 5.5.1 Not implemented.
-#500 5.5.1 Not implemented.^M
-#mailfront[#]: . 
-#mailfront[#]: 500 5.5.1 Not implemented.
-#500 5.5.1 Not implemented.^M
-#mailfront[#]: bytes in: 71 bytes out: 243
-#queuedir tmp: 0 new: 0
-#END_OF_TEST_RESULTS
-#
-#vecho "Running test tests/plugin-rbl '127.0.0.2' 'true'"
-#run_compare_test tests/plugin-rbl '127.0.0.2' 'true' <<END_OF_TEST_RESULTS
-#mailfront[#]: rbl: sbl-xbl.spamhaus.org:
-#mailfront[#]: http://www.spamhaus.org/sbl/query/SBL233
-#mailfront[#]: http://www.spamhaus.org/query/bl?ip=127.0.0.2
-#250 local.host^M
-#mailfront[#]: MAIL FROM:<somewhere>
-#250 Sender='somewhere'.^M
-#mailfront[#]: RCPT TO:<elsewhere>
-#250 Recipient='elsewhere'.^M
-#354 End your message with a period on a line by itself.^M
-#mailfront[#]: Subject: test
-#mailfront[#]: 451 Blocked: http://www.spamhaus.org/sbl/query/SBL233
-#451 Blocked: http://www.spamhaus.org/sbl/query/SBL233^M
-#mailfront[#]: bytes in: 71 bytes out: 213
-#queuedir tmp: 0 new: 1
-#===== queued file =====
-#somewhere^@elsewhere^@^@Subject: test
-#=====
-#END_OF_TEST_RESULTS
-#
-#vecho "Running test tests/plugin-rbl '127.0.0.99' 'false'"
-#run_compare_test tests/plugin-rbl '127.0.0.99' 'false' <<END_OF_TEST_RESULTS
-#250 local.host^M
-#mailfront[#]: MAIL FROM:<somewhere>
-#250 Sender='somewhere'.^M
-#mailfront[#]: RCPT TO:<elsewhere>
-#250 Recipient='elsewhere'.^M
-#354 End your message with a period on a line by itself.^M
-#mailfront[#]: Subject: test
-#250 Received 14 bytes.^M
-#mailfront[#]: bytes in: 71 bytes out: 182
-#queuedir tmp: 0 new: 0
-#END_OF_TEST_RESULTS
-#
-#vecho "Running test tests/plugin-rbl '127.0.0.99' 'true'"
-#run_compare_test tests/plugin-rbl '127.0.0.99' 'true' <<END_OF_TEST_RESULTS
-#250 local.host^M
-#mailfront[#]: MAIL FROM:<somewhere>
-#250 Sender='somewhere'.^M
-#mailfront[#]: RCPT TO:<elsewhere>
-#250 Recipient='elsewhere'.^M
-#354 End your message with a period on a line by itself.^M
-#mailfront[#]: Subject: test
-#250 Received 14 bytes.^M
-#mailfront[#]: bytes in: 71 bytes out: 182
-#queuedir tmp: 0 new: 0
-#END_OF_TEST_RESULTS
-#
+runtest() {
+local ip="$1"
+local queuedir="$2"
+
+export RBL_BLACKLISTS=zen.spamhaus.org
+export RBL_DEBUG=1
+export TCPREMOTEIP=$ip
+
+setup_queuedir
+if $queuedir
+then
+    export RBL_QUEUEDIR=$QUEUEDIR
+fi
+
+smtpfront echo rbl accept <<EOF
+HELO me
+MAIL FROM:<somewhere>
+RCPT TO:<elsewhere>
+DATA
+Subject: test
+.
+EOF
+
+dump_queuedir
+
+cleanup_queuedir
+
+unset RBL_BLACKLISTS RBL_DEBUG RBL_QUEUEDIR TCPLOCALIP QUEUEDIR
+rm -rf $tmp/queuedir $tmp/queuedir
+}
+vecho "Running test tests/plugin-rbl '127.0.0.2' 'false'"
+run_compare_test tests/plugin-rbl '127.0.0.2' 'false' <<END_OF_TEST_RESULTS
+mailfront[#]: rbl: blacklisted by zen.spamhaus.org: https://www.spamhaus.org/sbl/query/SBL2 // https://www.spamhaus.org/query/ip/127.0.0.2
+250 local.host^M
+mailfront[#]: MAIL FROM:<somewhere>
+mailfront[#]: 451 Blocked: https://www.spamhaus.org/sbl/query/SBL2
+451 Blocked: https://www.spamhaus.org/sbl/query/SBL2^M
+mailfront[#]: RCPT TO:<elsewhere>
+mailfront[#]: 503 5.5.1 You must send MAIL FROM: first
+503 5.5.1 You must send MAIL FROM: first^M
+mailfront[#]: 503 5.5.1 You must send MAIL FROM: first
+503 5.5.1 You must send MAIL FROM: first^M
+mailfront[#]: Subject: test
+mailfront[#]: 500 5.5.1 Not implemented.
+500 5.5.1 Not implemented.^M
+mailfront[#]: . 
+mailfront[#]: 500 5.5.1 Not implemented.
+500 5.5.1 Not implemented.^M
+mailfront[#]: bytes in: 71 bytes out: 242
+queuedir tmp: 0 new: 0
+END_OF_TEST_RESULTS
+
+vecho "Running test tests/plugin-rbl '127.0.0.2' 'true'"
+run_compare_test tests/plugin-rbl '127.0.0.2' 'true' <<END_OF_TEST_RESULTS
+mailfront[#]: rbl: blacklisted by zen.spamhaus.org: https://www.spamhaus.org/sbl/query/SBL2 // https://www.spamhaus.org/query/ip/127.0.0.2
+250 local.host^M
+mailfront[#]: MAIL FROM:<somewhere>
+250 Sender='somewhere'.^M
+mailfront[#]: RCPT TO:<elsewhere>
+250 Recipient='elsewhere'.^M
+354 End your message with a period on a line by itself.^M
+mailfront[#]: Subject: test
+mailfront[#]: 451 Blocked: https://www.spamhaus.org/sbl/query/SBL2
+451 Blocked: https://www.spamhaus.org/sbl/query/SBL2^M
+mailfront[#]: bytes in: 71 bytes out: 212
+queuedir tmp: 0 new: 1
+===== queued file =====
+somewhere^@elsewhere^@^@Subject: test
+=====
+END_OF_TEST_RESULTS
+
+vecho "Running test tests/plugin-rbl '127.0.0.99' 'false'"
+run_compare_test tests/plugin-rbl '127.0.0.99' 'false' <<END_OF_TEST_RESULTS
+250 local.host^M
+mailfront[#]: MAIL FROM:<somewhere>
+250 Sender='somewhere'.^M
+mailfront[#]: RCPT TO:<elsewhere>
+250 Recipient='elsewhere'.^M
+354 End your message with a period on a line by itself.^M
+mailfront[#]: Subject: test
+250 Received 14 bytes.^M
+mailfront[#]: bytes in: 71 bytes out: 182
+queuedir tmp: 0 new: 0
+END_OF_TEST_RESULTS
+
+vecho "Running test tests/plugin-rbl '127.0.0.99' 'true'"
+run_compare_test tests/plugin-rbl '127.0.0.99' 'true' <<END_OF_TEST_RESULTS
+250 local.host^M
+mailfront[#]: MAIL FROM:<somewhere>
+250 Sender='somewhere'.^M
+mailfront[#]: RCPT TO:<elsewhere>
+250 Recipient='elsewhere'.^M
+354 End your message with a period on a line by itself.^M
+mailfront[#]: Subject: test
+250 Received 14 bytes.^M
+mailfront[#]: bytes in: 71 bytes out: 182
+queuedir tmp: 0 new: 0
+END_OF_TEST_RESULTS
+
 
 ##### Test tests/rules-empty #####
 
@@ -923,6 +982,17 @@ run_compare_test tests/plugin-lua  <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
+##### Test tests/smtp-greeting #####
+
+runtest() {
+env SMTPGREETING='hostname hello' $src/mailfront smtp echo 2>/dev/null </dev/null
+}
+vecho "Running test tests/smtp-greeting "
+run_compare_test tests/smtp-greeting  <<END_OF_TEST_RESULTS
+220 hostname hello ESMTP^M
+END_OF_TEST_RESULTS
+
+
 ##### Test tests/plugin-check-fqdn #####
 
 runtest() {
@@ -1229,11 +1299,11 @@ run_compare_test tests/rules-databytes  <<END_OF_TEST_RESULTS
 250-8BITMIME^M
 250-ENHANCEDSTATUSCODES^M
 250 PIPELINING^M
-250 OK^M
-250 OK^M
-250 OK^M
-250 OK^M
-250 OK^M
+250 Sender='a@example.com'.^M
+250 Sender='a@example.com'. [SIZE]^M
+250 Sender='a@example.com'. [SIZE=]^M
+250 Sender='a@example.com'. [SIZE=100]^M
+250 Sender='a@example.com'. [SIZE=123]^M
 552 5.2.3 The message would exceed the maximum message size.^M
 503 5.5.1 You must send MAIL FROM: first^M
 END_OF_TEST_RESULTS
@@ -1446,6 +1516,97 @@ cur/1000000000.12345.here:1fsd
 END_OF_TEST_RESULTS
 
 
+##### Test tests/plugin-debug #####
+
+runtest() {
+export MSG_DEBUG=1
+
+smtpfront echo check-fqdn accept-recipient accept-sender <<EOF
+EHLO $HOSTNAME
+MAIL FROM:<from>
+MAIL FROM:<from@example.com>
+RCPT TO:<rcpt@example.net>
+RCPT TO:<rcpt@example.org>
+DATA
+Subject: test
+
+test
+.
+EOF
+
+unset MSG_DEBUG
+}
+vecho "Running test tests/plugin-debug "
+run_compare_test tests/plugin-debug  <<END_OF_TEST_RESULTS
+mailfront[#]: Debug: Calling plugin hook: init
+mailfront[#]: Debug: Calling plugin hook: reset
+mailfront[#]: Debug: Calling plugin hook: helo
+250-local.host^M
+250-8BITMIME^M
+250-ENHANCEDSTATUSCODES^M
+250 PIPELINING^M
+mailfront[#]: MAIL FROM:<from>
+mailfront[#]: Debug: Calling plugin hook: reset
+mailfront[#]: Debug: Calling plugin hook: sender
+mailfront[#]: Debug: Calling plugin: check-fqdn
+mailfront[#]: Debug: Plugin check-fqdn returned 554:5.1.2 Address is missing a domain name
+mailfront[#]: 554 5.1.2 Address is missing a domain name
+554 5.1.2 Address is missing a domain name^M
+mailfront[#]: MAIL FROM:<from@example.com>
+mailfront[#]: Debug: Calling plugin hook: reset
+mailfront[#]: Debug: Calling plugin hook: sender
+mailfront[#]: Debug: Calling plugin: check-fqdn
+mailfront[#]: Debug: Calling plugin: accept-sender
+mailfront[#]: Debug: Plugin accept-sender returned 250:(null)
+mailfront[#]: Debug: Calling plugin: accept
+mailfront[#]: Debug: Plugin accept returned 250:(null)
+250 Sender='from@example.com'.^M
+mailfront[#]: RCPT TO:<rcpt@example.net>
+mailfront[#]: Debug: Calling plugin hook: recipient
+mailfront[#]: Debug: Calling plugin: check-fqdn
+mailfront[#]: Debug: Calling plugin: accept-recipient
+mailfront[#]: Debug: Plugin accept-recipient returned 250:(null)
+mailfront[#]: Debug: Calling plugin: accept
+mailfront[#]: Debug: Plugin accept returned 250:(null)
+250 Recipient='rcpt@example.net'.^M
+mailfront[#]: RCPT TO:<rcpt@example.org>
+mailfront[#]: Debug: Calling plugin hook: recipient
+mailfront[#]: Debug: Calling plugin: check-fqdn
+mailfront[#]: Debug: Calling plugin: accept-recipient
+mailfront[#]: Debug: Plugin accept-recipient returned 250:(null)
+mailfront[#]: Debug: Calling plugin: accept
+mailfront[#]: Debug: Plugin accept returned 250:(null)
+250 Recipient='rcpt@example.org'.^M
+mailfront[#]: Debug: Calling plugin hook: data_start
+354 End your message with a period on a line by itself.^M
+mailfront[#]: Subject: test
+mailfront[#]: Debug: Calling plugin hook: message_end
+250 Received 20 bytes.^M
+mailfront[#]: Debug: Calling plugin hook: reset
+mailfront[#]: bytes in: 133 bytes out: 330
+END_OF_TEST_RESULTS
+
+
+##### Test tests/starttls-immediate #####
+
+runtest() {
+setup_cert
+make_stunnel_conf 'smtp echo accept'
+
+env TLS_CERTFILE=$certfile TLS_KEYFILE=$keyfile TLS_IMMEDIATE=1 \
+stunnel $tmp/stunnel.conf < /dev/null 2>&1 | fixtls
+}
+vecho "Running test tests/starttls-immediate "
+run_compare_test tests/starttls-immediate  <<END_OF_TEST_RESULTS
+LOG5[#]: Service [stunnel] accepted connection
+mailfront[#]: Starting TLS handshake
+mailfront[#]: TLS handshake: TLS#.# ECDHE-RSA AES-256-GCM AEAD
+220 local.host mailfront ESMTP^M
+mailfront[#]: bytes in: 0 bytes out: 32
+LOG5[#]: Connection closed: 0 byte(s) sent to TLS, 32 byte(s) sent to socket
+END_OF_TEST_RESULTS
+
+
 ##### Test tests/imapfront-auth-toomany #####
 
 runtest() {
@@ -1662,44 +1823,6 @@ Yes: 3
 END_OF_TEST_RESULTS
 
 
-##### Test tests/smtpfront-content #####
-
-runtest() {
-RELAYCLIENT= sfecho <<EOF
-MAIL FROM:<user@example.net>
-RCPT TO:<user@example.com>
-DATA
-Subject: test
-
-foo
-..
-bar
-.
-MAIL FROM:<user@example.net>
-RCPT TO:<user@example.com>
-DATA
-Subject: test
-
-foo
-..
-bar
-
-.
-EOF
-}
-vecho "Running test tests/smtpfront-content "
-run_compare_test tests/smtpfront-content  <<END_OF_TEST_RESULTS
-250 Sender='user@example.net'.^M
-250 Recipient='user@example.com'.^M
-354 End your message with a period on a line by itself.^M
-250 Received 25 bytes.^M
-250 Sender='user@example.net'.^M
-250 Recipient='user@example.com'.^M
-354 End your message with a period on a line by itself.^M
-250 Received 26 bytes.^M
-END_OF_TEST_RESULTS
-
-
 ##### Test tests/plugin-cvm-validate #####
 
 runtest() {
@@ -2003,19 +2126,6 @@ run_compare_test tests/pop3front-auth  <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
-##### Test tests/qmtpfront-echo #####
-
-runtest() {
-printf '6:\ntest\n,13:a@example.com,17:13:b@example.com,,' \
-| $src/mailfront qmtp echo 2>/dev/null
-echo
-}
-vecho "Running test tests/qmtpfront-echo "
-run_compare_test tests/qmtpfront-echo  <<END_OF_TEST_RESULTS
-18:KReceived 5 bytes.,
-END_OF_TEST_RESULTS
-
-
 ##### Test tests/pop3front-auth-userpass #####
 
 runtest() {
@@ -2204,6 +2314,48 @@ run_compare_test tests/rules-header-add  <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
+##### Test tests/rules-and #####
+
+runtest() {
+PLUGINS=mailrules:accept
+
+cat >$tmp/domains <<EOF
+example.com
+EOF
+
+cat >$tmp/rules <<EOF
+&a@*:*:K1
+k[[@$tmp/domains]]:*:K2
+&b@*:*:K3
+&*@example.com:*:K4
+k*:*:K5
+k*:*:K6
+EOF
+
+export MAILRULES=$tmp/rules
+
+sfecho <<EOF
+MAIL FROM:<a@example.com>
+MAIL FROM:<b@example.com>
+MAIL FROM:<c@example.com>
+MAIL FROM:<a@example.net>
+MAIL FROM:<b@example.net>
+MAIL FROM:<c@example.net>
+EOF
+
+rm -f $tmp/domains $tmp/rules
+}
+vecho "Running test tests/rules-and "
+run_compare_test tests/rules-and  <<END_OF_TEST_RESULTS
+250 K2^M
+250 K5^M
+250 K6^M
+250 K6^M
+250 K6^M
+250 K6^M
+END_OF_TEST_RESULTS
+
+
 ##### Test tests/pop3front-maildir-sort #####
 
 runtest() {
@@ -2226,6 +2378,88 @@ run_compare_test tests/pop3front-maildir-sort  <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
+##### Test tests/starttls-dhparams #####
+
+runtest() {
+setup_cert
+make_stunnel_conf 'smtp echo accept' 'protocol = smtp'
+env TLS_CERTFILE=$certfile TLS_KEYFILE=$keyfile TLS_DH_PARAMS=$dhparams \
+stunnel $tmp/stunnel.conf < /dev/null 2>&1 | fixtls
+}
+vecho "Running test tests/starttls-dhparams "
+run_compare_test tests/starttls-dhparams  <<END_OF_TEST_RESULTS
+LOG5[#]: Service [stunnel] accepted connection
+220 local.host mailfront ESMTP^M
+mailfront[#]: Starting TLS handshake
+mailfront[#]: TLS handshake: TLS#.# ECDHE-RSA AES-256-GCM AEAD
+mailfront[#]: bytes in: 0 bytes out: 0
+LOG5[#]: Connection closed: 0 byte(s) sent to TLS, 0 byte(s) sent to socket
+END_OF_TEST_RESULTS
+
+
+##### Test tests/smtp-addrparse #####
+
+runtest() {
+sfecho <<EOF
+MAIL FROM
+MAIL FROM 
+MAIL FROM:
+MAIL FROM<
+MAIL FROM>
+MAIL FROM me
+MAIL FROM:me
+MAIL FROM<me>
+MAIL FROM:<me>
+MAIL FROM me PARAM=VALUE
+MAIL FROM:me PARAM=VALUE
+MAIL FROM<me> PARAM=VALUE
+MAIL FROM:<me> PARAM=VALUE
+RCPT TO
+RCPT TO 
+RCPT TO:
+RCPT TO<
+RCPT TO>
+RCPT TO you
+RCPT TO:you
+RCPT TO<you>
+RCPT TO:<you>
+RCPT TO you PARAM=VALUE
+RCPT TO:you PARAM=VALUE
+RCPT TO<you> PARAM=VALUE
+RCPT TO:<you> PARAM=VALUE
+EOF
+}
+vecho "Running test tests/smtp-addrparse "
+run_compare_test tests/smtp-addrparse  <<END_OF_TEST_RESULTS
+501 5.5.2 Syntax error in address parameter.^M
+501 5.5.2 Syntax error in address parameter.^M
+501 5.5.2 Syntax error in address parameter.^M
+501 5.5.2 Syntax error in address parameter.^M
+501 5.5.2 Syntax error in address parameter.^M
+250 Sender='me'.^M
+250 Sender='me'.^M
+250 Sender='me'.^M
+250 Sender='me'.^M
+250 Sender='me'. [PARAM=VALUE]^M
+250 Sender='me'. [PARAM=VALUE]^M
+250 Sender='me'. [PARAM=VALUE]^M
+250 Sender='me'. [PARAM=VALUE]^M
+501 5.5.2 Syntax error in address parameter.^M
+501 5.5.2 Syntax error in address parameter.^M
+501 5.5.2 Syntax error in address parameter.^M
+501 5.5.2 Syntax error in address parameter.^M
+501 5.5.2 Syntax error in address parameter.^M
+250 Recipient='you'.^M
+250 Recipient='you'.^M
+250 Recipient='you'.^M
+250 Recipient='you'.^M
+250 Recipient='you'. [PARAM=VALUE]^M
+250 Recipient='you'. [PARAM=VALUE]^M
+250 Recipient='you'. [PARAM=VALUE]^M
+250 Recipient='you'. [PARAM=VALUE]^M
+END_OF_TEST_RESULTS
+
+
 ##### Test tests/imapfront-auth #####
 
 runtest() {
@@ -2266,14 +2500,80 @@ Yes: 7
 END_OF_TEST_RESULTS
 
 
-##### Test tests/smtpgreeting #####
+##### Test tests/smtp-content #####
 
 runtest() {
-env SMTPGREETING='hostname hello' $src/mailfront smtp echo 2>/dev/null </dev/null
+RELAYCLIENT= sfecho <<EOF
+MAIL FROM:<user@example.net>
+RCPT TO:<user@example.com>
+DATA
+Subject: test
+
+foo
+..
+bar
+.
+MAIL FROM:<user@example.net>
+RCPT TO:<user@example.com>
+DATA
+Subject: test
+
+foo
+..
+bar
+
+.
+EOF
 }
-vecho "Running test tests/smtpgreeting "
-run_compare_test tests/smtpgreeting  <<END_OF_TEST_RESULTS
-220 hostname hello ESMTP^M
+vecho "Running test tests/smtp-content "
+run_compare_test tests/smtp-content  <<END_OF_TEST_RESULTS
+250 Sender='user@example.net'.^M
+250 Recipient='user@example.com'.^M
+354 End your message with a period on a line by itself.^M
+250 Received 25 bytes.^M
+250 Sender='user@example.net'.^M
+250 Recipient='user@example.com'.^M
+354 End your message with a period on a line by itself.^M
+250 Received 26 bytes.^M
+END_OF_TEST_RESULTS
+
+
+##### Test tests/received-auth #####
+
+runtest() {
+local add_received="$1"
+PLUGINS=add-received:accept
+TESTOCALHOST="localhost"
+TESTLOCALIP="localip"
+TESTREMOTEHOST="remotehost"
+TESTREMOTEIP="remoteip"
+AUTH_ADD_RECEIVED=$add_received
+RELAYCLIENT=1
+
+export AUTH_ADD_RECEIVED RELAYCLIENT
+
+$src/mailfront smtp echo 2>&1 >/dev/null <<EOF | sed -n -e 's/^.* Received: //p'
+$helo
+MAIL FROM:<>
+RCPT TO:<test@example.com>
+DATA
+.
+EOF
+
+unset AUTH_ADD_RECEIVED RELAYCLIENT
+}
+vecho "Running test tests/received-auth ''"
+run_compare_test tests/received-auth '' <<END_OF_TEST_RESULTS
+from remotehost (remotehost [remoteip])  by local.host ([localip])
+END_OF_TEST_RESULTS
+
+vecho "Running test tests/received-auth '0'"
+run_compare_test tests/received-auth '0' <<END_OF_TEST_RESULTS
+END_OF_TEST_RESULTS
+
+vecho "Running test tests/received-auth '1'"
+run_compare_test tests/received-auth '1' <<END_OF_TEST_RESULTS
+from remotehost (remotehost [remoteip])  by local.host ([localip])
 END_OF_TEST_RESULTS
 
 
@@ -2375,14 +2675,14 @@ run_compare_test tests/rules-databytes3  <<END_OF_TEST_RESULTS
 250-8BITMIME^M
 250-ENHANCEDSTATUSCODES^M
 250 PIPELINING^M
-250 OK^M
-250 OK^M
-250 OK^M
-250 OK^M
-250 OK^M
+250 Sender='a@example.com'.^M
+250 Sender='a@example.com'. [SIZE]^M
+250 Sender='a@example.com'. [SIZE=]^M
+250 Sender='a@example.com'. [SIZE=100]^M
+250 Sender='a@example.com'. [SIZE=123]^M
 552 5.2.3 The message would exceed the maximum message size.^M
 503 5.5.1 You must send MAIL FROM: first^M
-250 OK^M
+250 Sender='a@example.com'.^M
 250 Recipient='nobody@example.com'.^M
 354 End your message with a period on a line by itself.^M
 552 5.2.3 Sorry, that message exceeds the maximum message length.^M
@@ -2574,27 +2874,6 @@ run_compare_test tests/plugin-counters-looping-delivered-to  <<END_OF_TEST_RESUL
 END_OF_TEST_RESULTS
 
 
-##### Test tests/smtpfront-addrfail #####
-
-runtest() {
-sfecho <<EOF
-MAIL FROM
-MAIL FROM 
-MAIL FROM:
-MAIL FROM<
-MAIL FROM>
-EOF
-}
-vecho "Running test tests/smtpfront-addrfail "
-run_compare_test tests/smtpfront-addrfail  <<END_OF_TEST_RESULTS
-501 5.5.2 Syntax error in address parameter.^M
-501 5.5.2 Syntax error in address parameter.^M
-501 5.5.2 Syntax error in address parameter.^M
-501 5.5.2 Syntax error in address parameter.^M
-501 5.5.2 Syntax error in address parameter.^M
-END_OF_TEST_RESULTS
-
-
 ##### Test tests/plugin-counters-maxrcpts #####
 
 runtest() {
@@ -2648,19 +2927,6 @@ run_compare_test tests/plugin-counters-maxrcpts  <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
-##### Test tests/qmqpfront-echo #####
-
-runtest() {
-printf '42:5:test\n,13:a@example.com,13:b@example.com,,' \
-| $src/mailfront qmqp echo 2>/dev/null
-echo
-}
-vecho "Running test tests/qmqpfront-echo "
-run_compare_test tests/qmqpfront-echo  <<END_OF_TEST_RESULTS
-18:KReceived 5 bytes.,
-END_OF_TEST_RESULTS
-
-
 ##### Test tests/imapfront-mailenv #####
 
 runtest() {
@@ -2765,6 +3031,77 @@ run_compare_test tests/rules-list  <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
+##### Test tests/smtp-maxnotimpl #####
+
+runtest() {
+MAXNOTIMPL=1
+
+sfecho <<EOF
+a
+b
+c
+d
+EOF
+
+MAXNOTIMPL=0
+
+sfecho <<EOF
+a
+b
+c
+d
+EOF
+}
+vecho "Running test tests/smtp-maxnotimpl "
+run_compare_test tests/smtp-maxnotimpl  <<END_OF_TEST_RESULTS
+500 5.5.1 Not implemented.^M
+503-5.5.0 Too many unimplemented commands.^M
+503 5.5.0 Closing connection.^M
+500 5.5.1 Not implemented.^M
+500 5.5.1 Not implemented.^M
+500 5.5.1 Not implemented.^M
+500 5.5.1 Not implemented.^M
+END_OF_TEST_RESULTS
+
+
+##### Test tests/starttls-nocert #####
+
+runtest() {
+make_stunnel_conf 'smtp echo accept' 'protocol = smtp'
+env TLS_CERTFILE=$tmp/does/not/exist \
+stunnel $tmp/stunnel.conf < /dev/null 2>&1 | fixtls
+}
+vecho "Running test tests/starttls-nocert "
+run_compare_test tests/starttls-nocert  <<END_OF_TEST_RESULTS
+LOG5[#]: Service [stunnel] accepted connection
+mailfront[#]: TLS init failed: Error while reading file.
+220 local.host mailfront ESMTP^M
+mailfront[#]: 500 5.5.1 Not implemented.
+LOG3[#]: Remote server is not RFC 2487 compliant
+LOG5[#]: Connection reset: 0 byte(s) sent to TLS, 0 byte(s) sent to socket
+mailfront[#]: bytes in: 26 bytes out: 131
+END_OF_TEST_RESULTS
+
+
+##### Test tests/smtp-quotes #####
+
+runtest() {
+sfecho <<EOF
+MAIL FROM:<"me, myself, and I"@example.net>
+RCPT TO:<"you, yourself, and you"@example.com>
+RCPT TO:<him\,himself@example.com>
+RCPT TO:<@somewhere,@elsewhere:two@example.com>
+EOF
+}
+vecho "Running test tests/smtp-quotes "
+run_compare_test tests/smtp-quotes  <<END_OF_TEST_RESULTS
+250 Sender='me, myself, and I@example.net'.^M
+250 Recipient='you, yourself, and you@example.com'.^M
+250 Recipient='him,himself@example.com'.^M
+250 Recipient='two@example.com'.^M
+END_OF_TEST_RESULTS
+
+
 ##### Test tests/plugin-counters-maxmsgs #####
 
 runtest() {
@@ -2830,6 +3167,56 @@ run_compare_test tests/rules-multiline  <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
+##### Test tests/starttls-combined #####
+
+runtest() {
+setup_cert
+make_stunnel_conf 'smtp echo accept' 'protocol = smtp'
+cat $certfile $keyfile > $tmp/tlsboth.pem
+env TLS_CERTFILE=$tmp/tlsboth.pem \
+stunnel $tmp/stunnel.conf < /dev/null 2>&1 | fixtls
+rm -f $tmp/tlsboth.pem
+}
+vecho "Running test tests/starttls-combined "
+run_compare_test tests/starttls-combined  <<END_OF_TEST_RESULTS
+LOG5[#]: Service [stunnel] accepted connection
+220 local.host mailfront ESMTP^M
+mailfront[#]: Starting TLS handshake
+mailfront[#]: TLS handshake: TLS#.# ECDHE-RSA AES-256-GCM AEAD
+mailfront[#]: bytes in: 0 bytes out: 0
+LOG5[#]: Connection closed: 0 byte(s) sent to TLS, 0 byte(s) sent to socket
+END_OF_TEST_RESULTS
+
+
+##### Test tests/rules-final #####
+
+runtest() {
+PLUGINS=mailrules:reject
+export REJECT=Rejected
+
+cat >$tmp/rules <<EOF
+ka@example.com:*
+Kb@example.com:*
+EOF
+
+export MAILRULES=$tmp/rules
+
+sfecho <<EOF
+MAIL FROM:<a@example.com>
+MAIL FROM:<b@example.com>
+EOF
+
+rm -f $tmp/domains $tmp/rules
+
+unset MAILRULES REJECT
+}
+vecho "Running test tests/rules-final "
+run_compare_test tests/rules-final  <<END_OF_TEST_RESULTS
+451 Rejected^M
+250 Sender='b@example.com'.^M
+END_OF_TEST_RESULTS
+
+
 ##### Test tests/rules-databytes2 #####
 
 runtest() {
@@ -2877,7 +3264,7 @@ run_compare_test tests/rules-databytes2  <<END_OF_TEST_RESULTS
 250-ENHANCEDSTATUSCODES^M
 250 PIPELINING^M
 250 Sender='somebody@example.net'. [SIZE=10000]^M
-250 OK^M
+250 Recipient='a@example.com'.^M
 354 End your message with a period on a line by itself.^M
 250 Received 8 bytes.^M
 250-local.host^M
@@ -2886,8 +3273,8 @@ run_compare_test tests/rules-databytes2  <<END_OF_TEST_RESULTS
 250-ENHANCEDSTATUSCODES^M
 250 PIPELINING^M
 250 Sender='somebody@example.net'. [SIZE=10000]^M
-250 OK^M
-250 OK^M
+250 Recipient='a@example.com'.^M
+250 Recipient='b@example.com'.^M
 354 End your message with a period on a line by itself.^M
 552 5.2.3 Sorry, that message exceeds the maximum message length.^M
 250-local.host^M
@@ -2896,47 +3283,14 @@ run_compare_test tests/rules-databytes2  <<END_OF_TEST_RESULTS
 250-ENHANCEDSTATUSCODES^M
 250 PIPELINING^M
 250 Sender='somebody@example.net'. [SIZE=10000]^M
-250 OK^M
-250 OK^M
-250 OK^M
+250 Recipient='a@example.com'.^M
+250 Recipient='b@example.com'.^M
+250 Recipient='c@example.com'.^M
 354 End your message with a period on a line by itself.^M
 552 5.2.3 Sorry, that message exceeds the maximum message length.^M
 END_OF_TEST_RESULTS
 
 
-##### Test tests/smtpfront-maxnotimpl #####
-
-runtest() {
-MAXNOTIMPL=1
-
-sfecho <<EOF
-a
-b
-c
-d
-EOF
-
-MAXNOTIMPL=0
-
-sfecho <<EOF
-a
-b
-c
-d
-EOF
-}
-vecho "Running test tests/smtpfront-maxnotimpl "
-run_compare_test tests/smtpfront-maxnotimpl  <<END_OF_TEST_RESULTS
-500 5.5.1 Not implemented.^M
-503-5.5.0 Too many unimplemented commands.^M
-503 5.5.0 Closing connection.^M
-500 5.5.1 Not implemented.^M
-500 5.5.1 Not implemented.^M
-500 5.5.1 Not implemented.^M
-500 5.5.1 Not implemented.^M
-END_OF_TEST_RESULTS
-
-
 ##### Test tests/rules-noop #####
 
 runtest() {
@@ -3116,6 +3470,19 @@ run_compare_test tests/rules-both  <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
+##### Test tests/qmqp-echo #####
+
+runtest() {
+printf '42:5:test\n,13:a@example.com,13:b@example.com,,' \
+| $src/mailfront qmqp echo 2>/dev/null
+echo
+}
+vecho "Running test tests/qmqp-echo "
+run_compare_test tests/qmqp-echo  <<END_OF_TEST_RESULTS
+18:KReceived 5 bytes.,
+END_OF_TEST_RESULTS
+
+
 ##### Test tests/pop3front-maildir-last #####
 
 runtest() {
@@ -3334,6 +3701,26 @@ run_compare_test tests/rules-maxhops  <<END_OF_TEST_RESULTS
 END_OF_TEST_RESULTS
 
 
+##### Test tests/starttls-baddhparams #####
+
+runtest() {
+setup_cert
+make_stunnel_conf 'smtp echo accept' 'protocol = smtp'
+env TLS_CERTFILE=$certfile TLS_KEYFILE=$keyfile TLS_DH_PARAMS=$tmp/does/not/exist \
+stunnel $tmp/stunnel.conf < /dev/null 2>&1 | fixtls
+}
+vecho "Running test tests/starttls-baddhparams "
+run_compare_test tests/starttls-baddhparams  <<END_OF_TEST_RESULTS
+LOG5[#]: Service [stunnel] accepted connection
+mailfront[#]: TLS error reading DH params: No such file or directory
+220 local.host mailfront ESMTP^M
+mailfront[#]: 500 5.5.1 Not implemented.
+LOG3[#]: Remote server is not RFC 2487 compliant
+LOG5[#]: Connection reset: 0 byte(s) sent to TLS, 0 byte(s) sent to socket
+mailfront[#]: bytes in: 26 bytes out: 131
+END_OF_TEST_RESULTS
+
+
 ##### Test tests/plugin-cvm-auth-plain #####
 
 runtest() {
@@ -3424,10 +3811,57 @@ vecho "Running test tests/rules-defaultmsg "
 run_compare_test tests/rules-defaultmsg  <<END_OF_TEST_RESULTS
 553 Rejected^M
 451 Deferred^M
-250 OK^M
+250 Sender='k@example.com'.^M
 553 Rejected^M
 451 Deferred^M
-250 OK^M
+250 Recipient='k@example.com'.^M
+END_OF_TEST_RESULTS
+
+
+##### Test tests/smtp-bad-bounce #####
+
+runtest() {
+# Note: this test no longer tests anything significant.
+
+sfecho <<EOF
+MAIL FROM:<notbounce@example.com>
+RCPT TO:<addr1@example.com>
+RCPT TO:<addr2@example.com>
+DATA
+.
+EOF
+
+sfecho <<EOF
+MAIL FROM:<>
+RCPT TO:<addr1@example.com>
+DATA
+.
+EOF
+
+sfecho <<EOF
+MAIL FROM:<>
+RCPT TO:<addr1@example.com>
+RCPT TO:<addr2@example.com>
+DATA
+.
+EOF
+}
+vecho "Running test tests/smtp-bad-bounce "
+run_compare_test tests/smtp-bad-bounce  <<END_OF_TEST_RESULTS
+250 Sender='notbounce@example.com'.^M
+250 Recipient='addr1@example.com'.^M
+250 Recipient='addr2@example.com'.^M
+354 End your message with a period on a line by itself.^M
+250 Received 0 bytes.^M
+250 Sender=''.^M
+250 Recipient='addr1@example.com'.^M
+354 End your message with a period on a line by itself.^M
+250 Received 0 bytes.^M
+250 Sender=''.^M
+250 Recipient='addr1@example.com'.^M
+250 Recipient='addr2@example.com'.^M
+354 End your message with a period on a line by itself.^M
+250 Received 0 bytes.^M
 END_OF_TEST_RESULTS
 
 
